/* 
 * packet sending and receiving routines for ncovert2
 *
 * $Id: packet.c,v 1.1.1.1 2004/03/17 16:55:30 s-nomad Exp $
 *
 * $Log: packet.c,v $
 * Revision 1.1.1.1  2004/03/17 16:55:30  s-nomad
 * Initial import for ncovert2-1.1
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <signal.h>
#include "ncovert.h"
#include "rand_gen.h"

unsigned int gen_rand_src( void )
{
	KEY saddr;
	int i;

	while(1)
	{
		for(i = 0; i<4; i++) saddr.c[i] = get_rand_num(R_BYTE);
		if((saddr.c[0] != 10) && 
		   (saddr.c[0] != 255) &&
		   (saddr.c[1] != 255) &&
		   (saddr.c[2] != 255) &&
		   (saddr.c[3] != 255) &&
		   (saddr.c[0] != 0) &&
		   (saddr.c[3] != 0) &&
		   (saddr.c[0] != 127) &&
		  ((saddr.c[0] != 192) && (saddr.c[0] != 168)))
			break;
	}
	return saddr.l;
}

void do_a_packet(unsigned short id, unsigned int source_addr, unsigned int dest_addr, unsigned int sport, unsigned int dport, unsigned long isn)
{
	struct send_tcp
	{
		struct iphdr ip;
		struct tcphdr tcp;
	} send_tcp;
	struct pseudo_header
	{
		unsigned int source_address;
		unsigned int dest_address;
		unsigned char placeholder;
		unsigned char protocol;
		unsigned short tcp_length;
		struct tcphdr tcp;
	} pseudo_header;
	struct sockaddr_in sai;
	int send_socket;
	
	send_tcp.ip.ihl = 5;
	send_tcp.ip.version = 4;
	send_tcp.ip.tos = 0;
	send_tcp.ip.tot_len = htons(40);
	send_tcp.ip.id = id;
	send_tcp.ip.frag_off = htons(0);
	send_tcp.ip.ttl = 64;
	send_tcp.ip.protocol = IPPROTO_TCP;
	send_tcp.ip.check = 0;
	send_tcp.ip.saddr = source_addr;
	send_tcp.ip.daddr = dest_addr;
	send_tcp.tcp.source = htons(sport);
	send_tcp.tcp.dest = htons(dport);
	send_tcp.tcp.seq = isn;
	send_tcp.tcp.ack_seq = 0;
	send_tcp.tcp.res1 = 0;
	send_tcp.tcp.doff = 5;
	send_tcp.tcp.fin = 0;
	send_tcp.tcp.syn = 1;
	send_tcp.tcp.rst = 0;
	send_tcp.tcp.psh = 0;
	send_tcp.tcp.ack = 0;
	send_tcp.tcp.urg = 0;
	send_tcp.tcp.window = htons(512);
	send_tcp.tcp.check = 0;
	send_tcp.tcp.urg_ptr = 0;

	sai.sin_family = AF_INET;
	sai.sin_port = send_tcp.tcp.source;
	sai.sin_addr.s_addr = send_tcp.ip.daddr;

	send_socket = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if(send_socket < 0)
	{
		perror("Unable to open socket");
		exit(1);
	}
	send_tcp.ip.check = in_cksum((unsigned short *)&send_tcp.ip, 20);
	pseudo_header.source_address = send_tcp.ip.saddr;
	pseudo_header.dest_address = send_tcp.ip.daddr;
	pseudo_header.placeholder = 0;
	pseudo_header.protocol = IPPROTO_TCP;
	pseudo_header.tcp_length = htons(20);
	bcopy((char *)&send_tcp.tcp, (char *)&pseudo_header.tcp, 20);
	send_tcp.tcp.check = in_cksum((unsigned short *)&pseudo_header, 32);
	sendto(send_socket, &send_tcp, 40, 0, (struct sockaddr *)&sai, sizeof(sai));
	close(send_socket);
}

void send_packets(unsigned int source_addr, unsigned int dest_addr, char *filename, int quiet) 
{
	FILE *input;
	blocks block;
	struct stat statbuf; 
	SI temp_s_port;
	unsigned short temp_short;
	unsigned long tempF;
	int size=0,extra,chunks,i,j=0,not_first=0,short_j=0;
	
	if(stat(filename,&statbuf) == -1)
	{
		printf("I cannot find the file %s for reading\n",filename);
		exit(1);
	}
	extra = (int)statbuf.st_size % 4;
	chunks = ((int)statbuf.st_size - extra) / 4; 
	if((input=fopen(filename,"rb"))== NULL)
	{
		printf("I cannot open the file %s for reading\n",filename);
		exit(1);
	}
	init_ncovert2(0);
	F = key.l ^ (unsigned long)statbuf.st_size;
	while(1)
	{
		if(!not_first) // do tcp ping first
		{
			do_a_packet(ipid.w, source_addr, dest_addr, s_port.w, 80, F); 
			temp_s_port = s_port;
			if (verbose) printf("sending data: 0x%08x\n",(unsigned int)F); 
			else if (!quiet) printf("."); 
			not_first = 1;
		}	
		else // sending of the data
		{
			for(i=0;i<chunks;i++)
			{
				guaranteed_memset(block.byte_chunk,0,4);
				size=fread(block.byte_chunk,1,4,input);
				if (size <= 0) 
				{
					printf("I cannot open the file %s for reading\n",filename);
					exit(1);
				}
				tempF = F ^ key.l;
				prepare_data(temp_s_port,block,j,tempF);
				temp_s_port = s_port;
				sleep(1);
				// randomly decide to send a decoy packet
				if(get_rand_num(R_BIT))
				{
					if(rand_saddr) source_addr = gen_rand_src();
					// first, gen a source port that is
					// NOT the one the server is looking
					// for...
					while(1)
					{
						temp_short = get_rand_num(R_UNSIGNEDWORD);
						if(temp_short != s_port.w) break;
					}
					// send the decoy with random values
					// and update the destination port
					do_a_packet(get_rand_num(R_UNSIGNEDWORD), source_addr, dest_addr, temp_short,dest_port[short_j],get_rand_num(R_UNSIGNEDLONG));
					short_j++;
					if(short_j == 65535) short_j = 0;
					sleep(1);
				}
				// send a real packet
				if(rand_saddr) source_addr = gen_rand_src();
				do_a_packet(get_rand_num(R_UNSIGNEDWORD), source_addr, dest_addr, s_port.w, dest_port[short_j], F);
				if (verbose) printf("sending data: 0x%08x (XOR'd as 0x%08x)\n",block.long_chunk,(unsigned int)F);
				else if (!quiet) printf(".");
				short_j++;
				if(short_j == 65535) short_j = 0;
				j = j + 2;
				if(j == 320) j = 0;
			}
			if (extra) // the last chunk of data...
			{
				guaranteed_memset(block.byte_chunk,0,4);
				size=fread(block.byte_chunk,1,4,input); 
				if (size <= 0)
				{
					printf("cannot open the file %s for reading\n",filename);
					exit(1);
				}
				tempF = F ^ key.l;
				prepare_data(temp_s_port,block,j,tempF);
				temp_s_port = s_port;
				sleep(1); 
				// send the real data
				if(rand_saddr) source_addr = gen_rand_src();
				do_a_packet(get_rand_num(R_UNSIGNEDWORD), source_addr, dest_addr, s_port.w, dest_port[short_j], F);
				if (verbose) printf("sending data: 0x%08x (XOR'd as 0x%08x)\n",block.long_chunk,(unsigned int)F); 
				else if (!quiet) printf("."); 
				// randomly decide to send a final decoy
				if(get_rand_num(R_BIT))
				{
					short_j++;
					if(short_j == 65535) short_j = 0;
					sleep(1);
					if(rand_saddr) source_addr = gen_rand_src();
					do_a_packet(get_rand_num(R_UNSIGNEDWORD), source_addr, dest_addr, get_rand_num(R_UNSIGNEDWORD),dest_port[short_j],get_rand_num(R_UNSIGNEDLONG));
				}
			}
			break;
		}
	}
	fclose(input);
}

void recv_packets(unsigned int source_addr, unsigned int dest_addr, char *filename, int quiet)
{ 
	struct recv_tcp
	{
		struct iphdr ip;
		struct tcphdr tcp;
		char buffer[10000];
	} recv_pkt;

	int recv_socket;
	FILE *output;
	blocks block;
	unsigned long tempF = 0;
	int filesize=0,counter=0,not_first=0,j=0;
	  
	if((output=fopen(filename,"wb"))== NULL)
	{ 
		printf("Can't open file %s\n",filename);
		exit(1);
	}
	init_ncovert2(1);
	signal(SIGALRM,caughtsig);
	alarm(timeout);
	while(1)
	{
		recv_socket = socket(AF_INET, SOCK_RAW, 6);
		if(recv_socket < 0)
		{
			perror("Unable to open socket");
			exit(1);
		}
		read(recv_socket, (struct recv_tcp *)&recv_pkt, 9999);
		  
		if(recv_pkt.ip.daddr == dest_addr)
		{
			if((not_first == 0) && 
			   (recv_pkt.tcp.syn==1) && 
			   (recv_pkt.ip.saddr == source_addr) && 
			   (ntohs(recv_pkt.tcp.dest) == 80))
			{ 
				alarm(0);
				s_port.w = ntohs(recv_pkt.tcp.source);
				init_server(recv_pkt.ip.id);
				if (verbose) printf("Receiving Data: 0x%08x\n",recv_pkt.tcp.seq);
				tempF = recv_pkt.tcp.seq;
				F = key.l ^ recv_pkt.tcp.seq;
				filesize = (int)F;
				counter = filesize;
				not_first++;
				determine_next_source_port(s_port,j);
				j = j + 2;
				if(j == 320) j = 0;
				alarm(timeout);
			}
			else if((ntohs(recv_pkt.tcp.source) == s_port.w) && 
				(recv_pkt.tcp.syn==1))
			{
				alarm(0);
				determine_next_source_port(s_port,j);
				j = j + 2;
				if(j == 320) j = 0;
				if (verbose) printf("Receiving Data: 0x%08x\n",recv_pkt.tcp.seq);
				F = tempF ^ recv_pkt.tcp.seq;
				block.long_chunk = key.l ^ F;
				tempF = recv_pkt.tcp.seq;
				if(counter < 4)
				{ 
					fwrite(block.byte_chunk,counter,1,output);
					break;
				} 
				else fwrite(block.byte_chunk,4,1,output);
				counter = counter - 4;
				fflush(output);
				alarm(timeout);
			}
		}
		close(recv_socket);
	}
	fclose(output);
}
